@@ -45,8 +45,11 @@ char ConnectMethod=CM_SERVER;
 struct SERVERLIST {
    char *Name;
    int Port;
+   int MaxPlayers,CurPlayers;
    char *Comment;
    char *Version;
+   char *Update;
+   char *UpSince;
    struct SERVERLIST *Next;
 };
 
@@ -67,7 +70,13 @@ int NumLocation,NumGun,NumDrug;
 struct PLAYER *Player,Noone;
 int LoanSharkLoc=DEFLOANSHARK,BankLoc=DEFBANK,GunShopLoc=DEFGUNSHOP,
     RoughPubLoc=DEFROUGHPUB;
-int DrugSortMethod=DS_ATOZ,FightTimeout=5,MaxClients=20,AITurnPause=5;
+char LoanSharkName[80]="the Loan Shark";
+char BankName[80]="the Bank";
+char GunShopName[80]="Dan\'s House of Guns";
+char RoughPubName[80]="the pub";
+int DrugSortMethod=DS_ATOZ;
+int FightTimeout=5,IdleTimeout=14400,ConnectTimeout=300;
+int MaxClients=20,AITurnPause=5;
 price_t StartCash=2000,StartDebt=5500;
 
 struct LOCATION StaticLocation,*Location;
@@ -85,7 +94,7 @@ struct BITCH Bitch = {
    50000,150000
 };
 struct METASERVER MetaServer = {
-   1,80,"bellatrix.pcl.ox.ac.uk","/~ben/cgi-bin/server.pl","","",
+   1,80,7802,"bellatrix.pcl.ox.ac.uk","/~ben/cgi-bin/server.pl","","",
    "dopewars server"
 };
 int NumTurns=31;
@@ -100,13 +109,17 @@ struct GLOBALS Globals[NUMGLOB] = {
    { &MetaServer.Active,NULL,NULL,-1,"MetaServer.Active",
      "Non-zero if server should report to a metaserver",
      NULL,NULL,0,"",NULL },
-   { &MetaServer.Port,NULL,NULL,-1,"MetaServer.Port",
-     "Port on which the metaserver listens",
+   { &MetaServer.HttpPort,NULL,NULL,-1,"MetaServer.HttpPort",
+     "Port used by the client to communicate with the metaserver",
+     NULL,NULL,0,"",NULL },
+   { &MetaServer.UdpPort,NULL,NULL,-1,"MetaServer.UdpPort",
+     "Port used by the server to communicate with the metaserver",
      NULL,NULL,0,"",NULL },
    { NULL,NULL,MetaServer.Name,160,"MetaServer.Name",
      "Metaserver name to report server details to",NULL,NULL,0,"",NULL },
    { NULL,NULL,MetaServer.Path,160,"MetaServer.Path",
-     "Path of the CGI script on the metaserver",NULL,NULL,0,"",NULL },
+     "Path of the CGI script, for the client, on the metaserver",
+     NULL,NULL,0,"",NULL },
    { NULL,NULL,MetaServer.LocalName,160,"MetaServer.LocalName",
      "Preferred hostname of your server machine",NULL,NULL,0,"",NULL },
    { NULL,NULL,MetaServer.Password,80,"MetaServer.Password",
@@ -136,12 +149,26 @@ struct GLOBALS Globals[NUMGLOB] = {
      NULL,NULL,0,"",NULL },
    { &RoughPubLoc,NULL,NULL,-1,"RoughPub","Location of the pub",
      NULL,NULL,0,"",NULL },
+   { NULL,NULL,LoanSharkName,80,"LoanSharkName",
+     "Name of the loan shark",NULL,NULL,0,"",NULL },
+   { NULL,NULL,BankName,80,"BankName",
+     "Name of the bank",NULL,NULL,0,"",NULL },
+   { NULL,NULL,GunShopName,80,"GunShopName",
+     "Name of the gun shop",NULL,NULL,0,"",NULL },
+   { NULL,NULL,RoughPubName,80,"RoughPubName",
+     "Name of the pub",NULL,NULL,0,"",NULL },
    { &DrugSortMethod,NULL,NULL,DS_MAX,"DrugSortMethod",
      "Sort key for listing available drugs",
      NULL,NULL,0,"",NULL },
    { &FightTimeout,NULL,NULL,-1,"FightTimeout",
      "No. of seconds in which to return fire",
      NULL,NULL,0,"",NULL },
+   { &IdleTimeout,NULL,NULL,-1,"IdleTimeout",
+     "Players are disconnected after this many seconds",
+     NULL,NULL,0,"",NULL },
+   { &ConnectTimeout,NULL,NULL,-1,"ConnectTimeout",
+     "Time in seconds for connections to be made or broken",
+     NULL,NULL,0,"",NULL },
    { &MaxClients,NULL,NULL,-1,"MaxClients",
      "Maximum number of TCP/IP connections",
      NULL,NULL,0,"",NULL },
@@ -416,10 +443,11 @@ int GetKey(char *allowed) {
 }
 
 int CountPlayers(struct PLAYER *First) {
-/* Returns the total numbers of players in the list starting at "First" */
+/* Returns the total numbers of players in the list starting at "First"; */
+/* players still in the process of connecting or leaving at ignored.     */
    struct PLAYER *tmp;
    int count=0;
-   for (tmp=First;tmp;tmp=tmp->Next) count++;
+   for (tmp=First;tmp;tmp=tmp->Next) if (tmp->Name[0]) count++;
    return count;
 }
 
@@ -437,6 +465,7 @@ struct PLAYER *AddPlayer(int fd,struct PLAYER **First) {
    NewPlayer->IsAt=0;
    NewPlayer->Attacked=NULL;
    NewPlayer->EventNum=E_NONE;
+   NewPlayer->FightTimeout=NewPlayer->ConnectTimeout=NewPlayer->IdleTimeout=0;
    NewPlayer->Guns=(Inventory *)calloc(NumGun,sizeof(Inventory));
    NewPlayer->Drugs=(Inventory *)calloc(NumDrug,sizeof(Inventory));
    if (!NewPlayer->Guns || !NewPlayer->Drugs) {
@@ -477,8 +506,8 @@ void RemovePlayer(struct PLAYER *Play,struct PLAYER **First) {
       if (!tmp) return;
       tmp->Next=Play->Next;
    }  
-   if (Server && Play->fd) {
-      close(Play->fd);
+   if (Server && Play->fd>=0) {
+      CloseSocket(Play->fd);
    }
    ClearList(&(Play->SpyList));
    ClearList(&(Play->TipList));
@@ -930,17 +959,19 @@ void DealDrugs(struct PLAYER *Play,char Buy) {
 price_t strtoprice(char *buf) {
 /* Forms a price based on the string representation in "buf"  */
    int i;
-   char digit,minus=0;
+   char digit,minus;
    price_t val=0;
+   minus=0;
    if (!buf) return 0;
    for (i=0;i<strlen(buf);i++) {
       digit=buf[i];
       if (digit>='0' && digit<='9') {
          val*=10;
          val+=(digit-'0');
-      } else if (digit='-' && i==0) minus=1;
+      } else if (digit=='-') minus=1;
    }
-   if (minus) return -val; else return val;
+   if (minus) val=-val;
+   return val;
 }
 
 char *pricetostr(price_t price,char *buf,int maxlen) {
@@ -948,21 +979,17 @@ char *pricetostr(price_t price,char *buf,int maxlen) {
 /* a pointer to it. If the string representation of the price */
 /* is longer than "maxlen" characters, it is truncated.       */
    int i;
-   char digit,minus=0;
+   char digit;
+   price_t absprice;
+   if (price<0) absprice=-price; else absprice=price;
    if (!buf) return NULL;
    memset(buf,0,maxlen);
-   if (price<0) {
-      price=-price;
-      minus=1;
-   }
    for (i=maxlen-2;i>=0;i--) {
-      digit='0'+(price%10);
-      price /= 10;
+      digit='0'+(absprice%10);
+      absprice /= 10;
       buf[i]=digit;
-      if (price==0) {
-         if (minus && i>0) {
-            buf[--i]='-';
-         }
+      if (absprice==0) {
+         if (price<0 && i>0) { i--; buf[i]='-'; }
          memmove(buf,&buf[i],maxlen-1-i);
          buf[maxlen-1-i]=0;
          return buf;
@@ -1476,12 +1503,12 @@ char *SelectServerFromMetaServer() {
    if ((he=gethostbyname(MetaServer.Name))==NULL) return NoHost;
    if ((HttpSock=socket(AF_INET,SOCK_STREAM,0))==-1) return NoSocket;
    HttpAddr.sin_family=AF_INET;
-   HttpAddr.sin_port=htons(MetaServer.Port);
+   HttpAddr.sin_port=htons(MetaServer.HttpPort);
    HttpAddr.sin_addr=*((struct in_addr *)he->h_addr);
    memset(HttpAddr.sin_zero,0,sizeof(HttpAddr.sin_zero));
    if (connect(HttpSock,(struct sockaddr *)&HttpAddr,
-       sizeof(struct sockaddr))==-1) {
-      close(HttpSock);
+       sizeof(struct sockaddr))==SOCKET_ERROR) {
+      CloseSocket(HttpSock);
       return NoService;
    }
    clear_line(17);
@@ -1490,7 +1517,7 @@ char *SelectServerFromMetaServer() {
    refresh();
    sprintf(buf,"GET %s?output=text&getlist=%d HTTP/1.0\n\n",
           MetaServer.Path,METAVERSION);
-   SendSocket(HttpSock,buf,strlen(buf));
+   send(HttpSock,buf,strlen(buf),0);
    ServerList=FALSE;
    while (bgets(buf,BUFLEN,HttpSock)) {
       chomp(buf); chomp(buf);
@@ -1501,15 +1528,22 @@ char *SelectServerFromMetaServer() {
          ThisServer->Port=atoi(buf);
          bgets(buf,BUFLEN,HttpSock); chomp(buf); chomp(buf);
          ThisServer->Version=strdup(buf);
-         bgets(buf,BUFLEN,HttpSock);
-         bgets(buf,BUFLEN,HttpSock);
+         bgets(buf,BUFLEN,HttpSock); chomp(buf); chomp(buf);
+         if (buf[0]) ThisServer->CurPlayers=atoi(buf);
+         else ThisServer->CurPlayers=-1;
+         bgets(buf,BUFLEN,HttpSock); chomp(buf); chomp(buf);
+         ThisServer->MaxPlayers=atoi(buf);
+         bgets(buf,BUFLEN,HttpSock); chomp(buf); chomp(buf);
+         ThisServer->Update=strdup(buf);
          bgets(buf,BUFLEN,HttpSock); chomp(buf); chomp(buf);
          ThisServer->Comment=strdup(buf);
+         bgets(buf,BUFLEN,HttpSock); chomp(buf); chomp(buf);
+         ThisServer->UpSince=strdup(buf);
          ThisServer->Next=FirstServer ? FirstServer : NULL;
          FirstServer=ThisServer;
       } else if (strncmp(buf,"MetaServer:",11)==0) ServerList=TRUE;
    }
-   close(HttpSock);
+   CloseSocket(HttpSock);
    ThisServer=FirstServer;
    while (ThisServer) {
       attrset(TextAttr);
@@ -1517,7 +1551,15 @@ char *SelectServerFromMetaServer() {
       mvaddstr(17,1,"Server : "); addstr(ThisServer->Name);
       sprintf(buf,"Port   : %d",ThisServer->Port);
       mvaddstr(18,1,buf);
-      mvaddstr(19,1,"Version: "); addstr(ThisServer->Version);
+      mvaddstr(18,40,"Version    : "); addstr(ThisServer->Version);
+      if (ThisServer->CurPlayers==-1) {
+         sprintf(buf,"Players: -unknown- (maximum %d)",ThisServer->MaxPlayers);
+      } else {
+         sprintf(buf,"Players: %d (maximum %d)",ThisServer->CurPlayers,
+                                                ThisServer->MaxPlayers);
+      }
+      mvaddstr(19,1,buf);
+      mvaddstr(19,40,"Up since   : "); addstr(ThisServer->UpSince);
       mvaddstr(20,1,"Comment: "); addstr(ThisServer->Comment);
       attrset(PromptAttr);
       mvaddstr(23,1,"N>ext server; P>revious server; S>elect this server... ");
@@ -1551,7 +1593,9 @@ char *SelectServerFromMetaServer() {
       NextServer=ThisServer->Next;
       free(ThisServer->Name);
       free(ThisServer->Version);
+      free(ThisServer->Update);
       free(ThisServer->Comment);
+      free(ThisServer->UpSince);
       free(ThisServer);
       ThisServer=NextServer;
    }
@@ -1869,8 +1913,9 @@ void do_game(struct PLAYER *Play) {
                   DisplayMode=DM_STREET; 
                   break;
                case 'R':
-                  DisplayMode=DM_NONE;
-                  jet(Play,FALSE);
+                  if (jet(Play,TRUE)) DisplayMode=DM_NONE;
+/*                  DisplayMode=DM_NONE;
+                  jet(Play,FALSE);*/
                   break;
                case 'F':
                   if (TotalGunsCarried(Play)>0 && Play->Flags&CANSHOOT) {
@@ -2277,12 +2322,9 @@ int main(int argc,char *argv[]) {
 /* options, before starting a client, server, or AI player.                */
    int c;
    char Name[NAMELEN];
-#if CYGWIN || HAVE_GETTIMEOFDAY
-   struct timeval tp;
 
-   gettimeofday(&tp,NULL);
-   srand(tp.tv_usec);
-#endif
+   srand(time(NULL));
+   PidFile=NULL;
    Location=NULL;
    Gun=NULL;
    Drug=NULL;
@@ -2293,7 +2335,7 @@ int main(int argc,char *argv[]) {
    WantColour=WantNetwork=1; WantHelp=WantVersion=WantAntique=0;
    Server=AIPlayer=FALSE;
    while (1) {
-      c=getopt(argc,argv,"anbchvf:o:sSp:g:");
+      c=getopt(argc,argv,"anbchvf:o:sSp:g:r:");
       if (c==EOF) break;
       switch(c) {
          case 'n': WantNetwork=0; break;
@@ -2309,6 +2351,7 @@ int main(int argc,char *argv[]) {
          case 'S': Server=TRUE; NotifyMetaServer=FALSE; break;
          case 'p': Port=atoi(optarg); break;
          case 'g': ProcessConfigFile(optarg); break;
+         case 'r': PidFile=strdup(optarg); break;
       }
    }
    if (WantVersion) {
@@ -2346,6 +2389,7 @@ dopewars metaserver)");
       puts("  -g file  specify the pathname of a dopewars configuration file. \
 This file is");
       puts("           read immediately when the -g option is encountered"); 
+      puts("  -r file  maintain pid file \"file\" while running the server");
       puts("  -c       create and run a computer player");
       puts("  -h       display this help information");
       puts("  -v       output version information and exit\n");
@@ -2397,5 +2441,6 @@ configure script.");
 #if NETWORKING
    StopNetworking();
 #endif
+   if (PidFile) free(PidFile);
    return 0;
 }
