@@ -1,5 +1,5 @@
 /* serverside.c   Handles the server side of dopewars                   */
-/* copyright (c)  1998-2000  Ben Webb                                   */
+/* copyright (c)  1998-2000  ben webb                                   */
 /*                Email: ben@bellatrix.pcl.ox.ac.uk                     */
 /*                WWW: http://bellatrix.pcl.ox.ac.uk/~ben/dopewars/     */
 
@@ -22,6 +22,7 @@
 #include <stdio.h>
 #include <string.h>
 #include <sys/types.h>
+#include <sys/stat.h>
 #ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #endif
@@ -37,7 +38,16 @@
 #include <fcntl.h>
 #endif
 
-int TerminateRequest;
+#define SD_SEND 1
+
+/* Maximum time, in seconds, between reports from this server to the   */
+/* metaserver. Setting this to be too small will most likely annoy the */
+/* metaserver maintainer, and result in your host being blocked... ;)  */
+#define METATIME (1200)
+
+int TerminateRequest,ReregisterRequest;
+
+int MetaTimeout;
 
 #define LOGF(x, args...) if (Network) printf(x, ## args)
 /* Prints a status message to standard output, unless the server is */
@@ -45,116 +55,79 @@ int TerminateRequest;
 
 struct PLAYER *FirstServer;
 
+/* Pointer to the filename of a pid file (if non-NULL) */
+char *PidFile;
+
 int SendSingleHighScore(struct PLAYER *Play,struct HISCORE *Score,
                         int index,char Bold);
 
-#if NETWORKING
-char *EncodeStringForCGI(char *str) {
-/* Takes the string "str" and replaces all spaces with "+" signs,  */
-/* and other non-alphabetic characters with %XX (where XX is the   */
-/* hexadecimal code of the character). This encoded string is then */
-/* suitable for passing to a CGI script; a dynamic buffer is       */
-/* allocated for this string, and returned. This buffer must be    */
-/* explicitly freed by the calling routine later. NULL is returned */
-/* if an error occurs.                                             */
-   char buf[20],*encode;
-   int StringLength,i;
-   if (!str) return NULL;
-   StringLength=strlen(str);
-   for (i=0;i<strlen(str);i++) {
-/* Test for non-alphabetic characters - do not use isalnum as this */
-/* may be locale-dependent */
-      if (str[i]!=' ' && !((str[i]>='a' && str[i]<='z') || 
-          (str[i]>='A' && str[i]<='Z') || (str[i]>='0' && str[i]<='9'))) {
-         StringLength+=2;
-      }
+int SendToMetaServer(char Up,int MetaSock,char *data,
+                     struct sockaddr_in *MetaAddr) {
+/* Sends server details, and any additional data, to the metaserver */
+   char buf[2000];
+   int numbytes;
+   sprintf(buf,"R:%d\n%d\n%s\n%s",
+           METAVERSION,Port,MetaServer.LocalName,MetaServer.Password);
+   if (data) { strcat(buf,"\n"); strcat(buf,data); }
+   numbytes=sendto(MetaSock,buf,strlen(buf),0,
+                   (struct sockaddr *)MetaAddr,sizeof(struct sockaddr));
+   if (numbytes==-1) {
+      printf("Error: cannot send data to metaserver\n");
+      return 0;
    }
-   encode=(char *)malloc(StringLength);
-   if (!encode) return NULL;
-   encode[0]=0;
-   for (i=0;i<strlen(str);i++) {
-      if (str[i]!=' ' && !((str[i]>='a' && str[i]<='z') || 
-          (str[i]>='A' && str[i]<='Z') || (str[i]>='0' && str[i]<='9'))) {
-         sprintf(buf,"%%%02X",str[i]); buf[3]=0;
-      } else if (str[i]==' ') {
-         buf[0]='+'; buf[1]=0;
-      } else { 
-         buf[0]=str[i]; buf[1]=0;
-      }
-      strcat(encode,buf);
-   }
-   return encode;
+   return 1;
 }
 
-void RegisterWithMetaServer(char Up) {
+void RegisterWithMetaServer(char Up,char SendData) {
 /* Sends server details to the metaserver, if specified. If "Up" is  */
 /* TRUE, informs the metaserver that the server is now accepting     */
 /* connections - otherwise tells the metaserver that this server is  */
-/* about to go down.                                                 */
-   struct sockaddr_in HttpAddr;
+/* about to go down. If "SendData" is TRUE, then also sends game     */
+/* data (e.g. scores) to the metaserver. If networking is disabled,  */
+/* does nothing.                                                     */
+#if NETWORKING
+   struct HISCORE MultiScore[NUMHISCORE],AntiqueScore[NUMHISCORE];
+   struct sockaddr_in MetaAddr;
    struct hostent *he;
-   int HttpSock;
-   char buf[900],report=FALSE;
-   char *Comment,*LocalName,*Password;
+   int MetaSock;
+   char buf[900],prstr[100];
+   int i,numbytes;
    if (!MetaServer.Active || !NotifyMetaServer) return;
-   printf("Attempting to establish connection with metaserver at %s\n",
-          MetaServer.Name);
+   printf("%s metaserver at %s\n",SendData ? "Sending data to" : "Notifying",
+                                  MetaServer.Name);
    if ((he=gethostbyname(MetaServer.Name))==NULL) {
       printf("Error: cannot locate metaserver\n");
       return;
    }
-   HttpSock=socket(AF_INET,SOCK_STREAM,0);
-   if (HttpSock==-1) {
+   MetaSock=socket(AF_INET,SOCK_DGRAM,0);
+   if (MetaSock==-1) {
       printf("Error: cannot create socket for metaserver communication\n");
       return;
    }
-   HttpAddr.sin_family=AF_INET;
-   HttpAddr.sin_port=htons(MetaServer.Port);
-   HttpAddr.sin_addr=*((struct in_addr *)he->h_addr);
-   memset(HttpAddr.sin_zero,0,sizeof(HttpAddr.sin_zero));
+   memset(&MetaAddr,0,sizeof(struct sockaddr_in));
+   MetaAddr.sin_family=AF_INET;
+   MetaAddr.sin_port=htons(MetaServer.UdpPort);
+   MetaAddr.sin_addr=*((struct in_addr *)he->h_addr);
 
-   if (connect(HttpSock,(struct sockaddr *)&HttpAddr,
-       sizeof(struct sockaddr))==-1) {
-      close(HttpSock);
-      printf("Error: Metaserver is not running an HTTP service or connection \
-denied\n");
-      return;
+   sprintf(buf,"report\n%d\n%s\n%d\n%d\n%s",
+           Up ? 1 : 0,VERSION,CountPlayers(FirstServer),
+           MaxClients,MetaServer.Comment);
+   if (!SendToMetaServer(Up,MetaSock,buf,&MetaAddr)) return;
+
+   if (SendData) {
+      if (HighScoreRead(MultiScore,AntiqueScore)) {
+         for (i=0;i<NUMHISCORE;i++) {
+            sprintf(buf,"multi\n%d\n%s^%s^%s^%c",
+                    i,FormatPrice(MultiScore[i].Money,prstr),MultiScore[i].Time,
+                    MultiScore[i].Name,MultiScore[i].Dead ? '1':'0');
+            if (!SendToMetaServer(Up,MetaSock,buf,&MetaAddr)) return;
+         }
+      } else { printf("Error: cannot read high score file\n"); }
    }
-   printf("Connection to metaserver established - sending details\n");
-   Comment=EncodeStringForCGI(MetaServer.Comment);
-   if (!Comment) {
-      printf("Error: cannot encode string for metaserver comment\n");   
-      return;
-   }
-   if (MetaServer.LocalName[0]) {
-      LocalName=EncodeStringForCGI(MetaServer.LocalName);
-      Password=EncodeStringForCGI(MetaServer.Password);
-      if (!LocalName || !Password) {
-         printf("Error: cannot encode strings for metaserver\n");
-         return;
-      }
-      sprintf(buf,"GET %s?output=text&status=%s&version=%s&port=%d\
-&players=%d&name=%s&password=%s&comment=%s HTTP/1.0\n\n",
-             MetaServer.Path,Up ? "up" : "down",VERSION,Port,
-             MaxClients,LocalName,Password,Comment);
-      free(LocalName); free(Password);
-   } else {
-      sprintf(buf,"GET %s?output=text&status=%s&version=%s&port=%d\
-&players=%d&comment=%s HTTP/1.0\n\n",
-             MetaServer.Path,Up ? "up" : "down",VERSION,Port,
-             MaxClients,Comment);
-   }
-   free(Comment);
-   SendSocket(HttpSock,buf,strlen(buf));
-   while (bgets(buf,900,HttpSock)) {
-      chomp(buf); chomp(buf);
-      if (report) puts(buf);
-      else if (strstr(buf,"text/")) report=TRUE;
-   }
-   close(HttpSock);
-   printf("Metaserver connection closed OK\n");
-}
+   CloseSocket(MetaSock);
+   MetaTimeout=time(NULL)+METATIME;
 #endif /* NETWORKING */
+}
 
 void SendGameData(struct PLAYER *To) {
 /* Sends all the game data required by a client (names of locations,  */
@@ -205,8 +178,9 @@ void HandleServerMessage(char *buf,struct PLAYER *ReallyFrom) {
       LOGF("Bad message\n");
       return;
    }
-   if (From!=ReallyFrom && (From!=&Noone || Code!=C_NAME)) {
-       LOGF("Message is lying about its orgin\n");
+   if (From!=ReallyFrom && (From!=&Noone ||
+       (Code!=C_NAME && Code!=C_NETMESSAGE))) {
+       LOGF("Message is lying about its origin\n");
        LOGF("%s: %c: %s: %s\n",From ? From->Name : "",Code,
                                  To ? To->Name : "",Data);
        LOGF("Should be from %s\n",ReallyFrom ? ReallyFrom->Name : "NULL");
@@ -214,27 +188,60 @@ void HandleServerMessage(char *buf,struct PLAYER *ReallyFrom) {
    }
    switch(Code) {
       case C_MSGTO:
-         LOGF("%s->%s: %c: %s\n",From->Name,To->Name,Code,Data);
+         LOGF("%s->%s: %s\n",From->Name,To->Name,Data);
          SendServerMessage(From,AICode,Code,To,Data);
          break;
+      case C_NETMESSAGE:
+         printf("Net:%s\n",Data);
+         shutdown(ReallyFrom->fd,SD_SEND);
+/* Make sure they do actually disconnect, eventually! */
+         if (ConnectTimeout) {
+            ReallyFrom->ConnectTimeout=time(NULL)+(time_t)ConnectTimeout;
+         }
+         break;
       case C_NAME:
          pt=GetPlayerByName(Data,FirstServer);
          if (strlen(Data)>NAMELEN || (pt && pt!=From)) {
+            if (ConnectTimeout) {
+               ReallyFrom->ConnectTimeout=time(NULL)+(time_t)ConnectTimeout;
+            }
             SendServerMessage(NULL,C_NONE,C_NEWNAME,ReallyFrom,NULL);
          } else if ((ReallyFrom && ReallyFrom->Name[0]==0 && Network) || 
                    (!Network && From==&Noone)) {
-            if (CountPlayers(FirstServer)<=MaxClients) { 
+            if (CountPlayers(FirstServer)<=MaxClients || !Network) { 
                SendGameData(ReallyFrom);
                if (!Network) From=AddPlayer(0,&FirstServer);
                else From=ReallyFrom;
                InitialisePlayer(From);
-            
+               ClearFightTimeout(From);
                strcpy(From->Name,Data);
+               for (pt=FirstServer;pt;pt=pt->Next) {
+                  if (pt!=From) {
+                     SendServerMessage(NULL,C_NONE,C_LIST,From,pt->Name);
+                  }
+               }
+               SendServerMessage(NULL,C_NONE,C_ENDLIST,From,NULL);
+               RegisterWithMetaServer(TRUE,FALSE);
+               From->ConnectTimeout=0;
+
                LOGF("%s joins the game!\n",From->Name);
                BroadcastToClients(C_NONE,C_JOIN,From->Name,NULL,From);
                From->EventNum=E_ARRIVE;
                SendPlayerData(From);
                SendEvent(From);
+            } else {
+               From=ReallyFrom;
+               printf("MaxClients (%d) exceeded - dropping connection\n",
+                      MaxClients);
+               sprintf(buf,"Sorry, but this server has a limit of %d \
+player%s, which has been reached.^Please try connecting again later.",
+                       MaxClients,MaxClients==1 ? "" : "s");
+               SendServerMessage(NULL,C_NONE,C_PRINTMESSAGE,From,buf);
+               shutdown(From->fd,SD_SEND);
+/* Make sure they do actually disconnect, eventually! */
+               if (ConnectTimeout) {
+                  From->ConnectTimeout=time(NULL)+(time_t)ConnectTimeout;
+               }
             }
          } else {
             LOGF("%s will now be known as %s\n",From->Name,Data);
@@ -378,6 +385,9 @@ int CheckServerSockets(fd_set *readfs,char WantQuit) {
 /* server to quit for good. Otherwise, 0 is returned.                */
    struct PLAYER *tmp;
    char buf[BUFLEN];
+   time_t timenow;
+
+   timenow=time(NULL);
    tmp=FirstServer;
    while (tmp) {
       if (FD_ISSET(tmp->fd,readfs)) {
@@ -385,11 +395,19 @@ int CheckServerSockets(fd_set *readfs,char WantQuit) {
             if (!WantQuit && tmp->Name[0]) {
                LOGF("%s leaves the server!\n",tmp->Name);
                ClientLeftServer(tmp);
+/* Blank the name, so that CountPlayers ignores this player */
+               tmp->Name[0]=0;
+/* Report the new high scores (if any) and the new number of players
+   to the metaserver */
+               RegisterWithMetaServer(TRUE,TRUE);
             }
             RemovePlayer(tmp,&FirstServer);
             if (!FirstServer && WantQuit) return -1;
             else return 0;
          } else {
+/* OK, we have some data. Reset the idle timeout (if necessary) and
+   handle the message */
+            if (IdleTimeout) tmp->IdleTimeout=timenow+(time_t)IdleTimeout;
             chomp(buf);
             HandleServerMessage(buf,tmp);
          }
@@ -402,17 +420,21 @@ int CheckServerSockets(fd_set *readfs,char WantQuit) {
 
 void CleanUpServer() {
 /* Closes down the server and frees up associated handles and memory */
-#if NETWORKING
-   if (Server) RegisterWithMetaServer(FALSE);
-#endif
+   if (Server) RegisterWithMetaServer(FALSE,FALSE);
    while (FirstServer) {
       RemovePlayer(FirstServer,&FirstServer);
    }
 #if NETWORKING
-   if (Server) close(ListenSock);
+   if (Server) CloseSocket(ListenSock);
 #endif
 }
 
+void ReregisterHandle(int sig) {
+/* Responds to a SIGUSR1 signal, and requests the main event loop to  */
+/* reregister the server with the dopewars metaserver.                */
+   ReregisterRequest=1;
+}
+
 void BreakHandle(int sig) {
 /* Traps an attempt by the user to send dopewars a SIGTERM or SIGINT  */
 /* (e.g. pressing Ctrl-C) and signals for a "nice" shutdown. Restores */
@@ -481,6 +503,25 @@ void ServerHelp() {
 }
 
 #if NETWORKING
+void CreatePidFile() {
+/* Creates a pid file (if "PidFile" is non-NULL) and writes the process */
+/* ID into it                                                           */
+   FILE *fp;
+   if (!PidFile) return;
+   fp=fopen(PidFile,"w");
+   if (fp) {
+      printf("Maintaining pid file %s\n",PidFile);
+      fprintf(fp,"%ld\n",(long)getpid());
+      fclose(fp);
+      chmod(PidFile,S_IREAD|S_IWRITE);
+   } else fprintf(stderr,"Cannot create pid file %s\n",PidFile);
+}
+
+void RemovePidFile() {
+/* Removes the previously-created pid file "PidFile" */
+   if (PidFile) unlink(PidFile);
+}
+
 void ServerLoop() {
 /* Initialises server, processes network and interactive messages, and */
 /* finally cleans up the server on exit.                               */
@@ -495,6 +536,8 @@ void ServerLoop() {
    struct sigaction sact;
    int MinTimeout;
 
+   CreatePidFile();
+
 /* Make the output line-buffered, so that the log file (if used) is */
 /* updated regularly                                                */
    fflush(stdout);
@@ -523,8 +566,6 @@ void ServerLoop() {
       perror("bind socket"); exit(1);
    }
 
-   RegisterWithMetaServer(TRUE);
-
    printf("dopewars server version %s ready and waiting for \
 connections on port %d\n\n",VERSION,Port);
    printf("For assistance with server commands, enter the command \"help\"\n");
@@ -533,7 +574,19 @@ connections on port %d\n\n",VERSION,Port);
       perror("listen socket"); exit(1);
    }
 
-   TerminateRequest=0;
+   MetaTimeout=0;
+
+   TerminateRequest=ReregisterRequest=0;
+
+#if !CYGWIN
+   sact.sa_handler=ReregisterHandle;
+   sact.sa_flags=0;
+   sigemptyset(&sact.sa_mask);
+   if (sigaction(SIGUSR1,&sact,NULL)==-1) {
+      printf("Cannot install SIGUSR1 interrupt handler!\n");
+   }
+#endif
+
    sact.sa_handler=BreakHandle;
    sact.sa_flags=0;
    sigemptyset(&sact.sa_mask);
@@ -543,11 +596,19 @@ connections on port %d\n\n",VERSION,Port);
    if (sigaction(SIGTERM,&sact,NULL)==-1) {
       printf("Cannot install SIGTERM interrupt handler!\n");
    }
+#if !CYGWIN
+   if (sigaction(SIGHUP,&sact,NULL)==-1) {
+      printf("Cannot install SIGHUP interrupt handler!\n");
+   }
+#endif
    sact.sa_handler=SIG_IGN;
    sact.sa_flags=0;
    if (sigaction(SIGPIPE,&sact,NULL)==-1) {
       printf("Cannot install pipe handler!\n");
    }
+
+   RegisterWithMetaServer(TRUE,TRUE);
+
    while (1) {
       FD_ZERO(&readfs);
       if (!InputClosed) FD_SET(0,&readfs);
@@ -556,9 +617,11 @@ connections on port %d\n\n",VERSION,Port);
       FD_SET(ListenSock,&errorfs);
       topsock=ListenSock+1;
       for (tmp=FirstServer;tmp;tmp=tmp->Next) {
-         FD_SET(tmp->fd,&readfs);
-         FD_SET(tmp->fd,&errorfs);
-         if (tmp->fd>=topsock) topsock=tmp->fd+1;
+         if (tmp->fd>0) {
+            FD_SET(tmp->fd,&readfs);
+            FD_SET(tmp->fd,&errorfs);
+            if (tmp->fd>=topsock) topsock=tmp->fd+1;
+         }
       }
       MinTimeout=GetMinimumTimeout(FirstServer);
       if (MinTimeout!=-1) {
@@ -568,11 +631,15 @@ connections on port %d\n\n",VERSION,Port);
       if (bselect(topsock,&readfs,NULL,&errorfs,
                  MinTimeout==-1 ? NULL : &timeout)==-1) {
          if (errno==EINTR) {
-            if (TerminateRequest) break; else continue;
+            if (ReregisterRequest) {
+               ReregisterRequest=0;
+               RegisterWithMetaServer(TRUE,TRUE);
+               continue;
+            } else if (TerminateRequest) break; else continue;
          }
          perror("select"); exit(1);
       }
-      HandleTimeouts(FirstServer);
+      HandleTimeouts(&FirstServer);
       if (FD_ISSET(0,&readfs)) {
          if (fgets(buf,sizeof(buf),stdin)==NULL) {
             buf[0]=0;
@@ -582,7 +649,7 @@ connections on port %d\n\n",VERSION,Port);
          chomp(buf);
          if (ParseNameValue(buf,name,80,value,80,sname,80,&Index,
              &HasEquals)==0 && !HandleOption(name,sname,Index,value,
-             "in command line",TRUE,FirstServer ? TRUE : FALSE,HasEquals)) {
+             "in command line",TRUE,CountPlayers(FirstServer) > 0,HasEquals)) {
             if (strcasecmp(name,"help")==0 || strcasecmp(name,"h")==0 ||
                 strcmp(name,"?")==0) {
                ServerHelp();
@@ -615,6 +682,9 @@ connections on port %d\n\n",VERSION,Port);
             }
          }
       }
+      if (CheckServerSockets(&readfs,WantQuit)==-1) {
+         break;
+      }
       if (FD_ISSET(ListenSock,&readfs)) {
          cadsize=sizeof(struct sockaddr);
          if ((ClientSock=accept(ListenSock,(struct sockaddr *)&ClientAddr,
@@ -624,27 +694,9 @@ connections on port %d\n\n",VERSION,Port);
          LOGF("server: got connection from %s\n",
                 inet_ntoa(ClientAddr.sin_addr));
          tmp=AddPlayer(ClientSock,&FirstServer);
-         if (tmp) {
-            if (CountPlayers(FirstServer)>MaxClients) {
-               printf("MaxClients (%d) exceeded - dropping connection\n",
-                      MaxClients);
-               sprintf(buf,"Sorry, but this server has a limit of %d \
-player%s, which has been reached.^Please try connecting again later.",
-                       MaxClients,MaxClients==1 ? "" : "s");
-               SendServerMessage(NULL,C_NONE,C_PRINTMESSAGE,tmp,buf);
-               shutdown(tmp->fd,2);
-            } else {
-               ClearFightTimeout(tmp);
-               for (tmp2=FirstServer;tmp2;tmp2=tmp2->Next) {
-                  if (tmp!=tmp2) {
-                     SendServerMessage(NULL,C_NONE,C_LIST,tmp,tmp2->Name);
-                  }
-               }
-               SendServerMessage(NULL,C_NONE,C_ENDLIST,tmp,NULL);
-            }
+         if (ConnectTimeout) {
+            tmp->ConnectTimeout=time(NULL)+(time_t)ConnectTimeout;
          }
-      } else if (CheckServerSockets(&readfs,WantQuit)==-1) {
-         break;
       }
       for (tmp=FirstServer;tmp;tmp=tmp->Next) if (FD_ISSET(tmp->fd,&errorfs)) {
          LOGF("error from client: %d\n",tmp->fd);
@@ -652,6 +704,7 @@ player%s, which has been reached.^Please try connecting again later.",
       }
    }
    CleanUpServer();
+   RemovePidFile();
 }
 #endif /* NETWORKING */
 
@@ -660,8 +713,11 @@ void FinishGame(struct PLAYER *Play,char *Message) {
    ClientLeftServer(Play);
    Play->EventNum=E_FINISH;
    SendHighScores(Play,TRUE,Message);
-   shutdown(Play->fd,2);
-   Play->Name[0]=0;
+   shutdown(Play->fd,SD_SEND);
+/* Make sure they do actually disconnect, eventually! */
+   if (ConnectTimeout) {
+      Play->ConnectTimeout=time(NULL)+(time_t)ConnectTimeout;
+   }
 }
 
 void HighScoreTypeRead(struct HISCORE *HiScore,FILE *fp) {
@@ -726,6 +782,8 @@ void SendHighScores(struct PLAYER *Play,char EndGame,char *Message) {
 /* if it's non-NULL. The client is then informed that the game's over. */
    struct HISCORE MultiScore[NUMHISCORE],AntiqueScore[NUMHISCORE],Score;
    struct HISCORE *HiScore;
+   struct tm *timep;
+   time_t tim;
    char text[BUFLEN];
    int i,j,InList=-1;
    if (!HighScoreRead(MultiScore,AntiqueScore)) {
@@ -741,7 +799,9 @@ void SendHighScores(struct PLAYER *Play,char EndGame,char *Message) {
       Score.Money=Play->Cash+Play->Bank-Play->Debt;
       strcpy(Score.Name,Play->Name);
       if (Play->Health==0) Score.Dead=1; else Score.Dead=0;
-      GetTime(Score.Time,80);
+      tim=time(NULL);
+      timep=gmtime(&tim);
+      strftime(Score.Time,80,"%d-%m-%Y",timep);
       for (i=0;i<NUMHISCORE;i++) {
          if (InList==-1 && (Score.Money > HiScore[i].Money ||
              HiScore[i].Time[0]==0)) {
@@ -875,30 +935,29 @@ discovered!^The spy %s!",To->Name,Discover[j]);
             break;
          case E_LOANSHARK:
             if (To->IsAt+1==LoanSharkLoc && To->Debt>0) {
-               SendQuestion(NULL,C_ASKLOAN,To,
-                            "YN^Would you like to visit the Loan Shark?");
+               sprintf(text,"YN^Would you like to visit %s?",LoanSharkName);
+               SendQuestion(NULL,C_ASKLOAN,To,text);
                return;
             }
             break;
          case E_BANK:
             if (To->IsAt+1==BankLoc) {
-               SendQuestion(NULL,C_ASKBANK,To,
-                                 "YN^Would you like to visit the Bank?");
+               sprintf(text,"YN^Would you like to visit %s?",BankName);
+               SendQuestion(NULL,C_ASKBANK,To,text);
                return;
             }
             break;
          case E_GUNSHOP:
             if (To->IsAt+1==GunShopLoc && !Sanitized && !WantAntique) {
-               sprintf(text,"YN^Would you like to visit Dan's House of %s?",
-                       InitialCaps(Names.Guns,buf));
+               sprintf(text,"YN^Would you like to visit %s?",GunShopName);
                SendQuestion(NULL,C_ASKGUNSHOP,To,text);
                return;
             }
             break;
          case E_ROUGHPUB:
             if (To->IsAt+1==RoughPubLoc && !WantAntique) {
-               SendQuestion(NULL,C_ASKPUB,To,
-                                 "YN^Would you like to visit the pub?");
+               sprintf(text,"YN^Would you like to visit %s?",RoughPubName);
+               SendQuestion(NULL,C_ASKPUB,To,text);
                return;
             }
             break;
@@ -1430,10 +1489,10 @@ trip you ever imagined!^Then you died because your brain disintegrated!");
                From->ResyncNum=E_ARRIVE+1;
                From->EventNum=E_ATTACK;
                AttackPlayer(From,From->Attacked,AT_FIRST | AT_SHOOT);
-            } else if (answer[0]=='T') {
+/*          } else if (answer[0]=='T') {
                From->Flags |= TRADING;
                SendPlayerData(From);
-               SendServerMessage(NULL,C_NONE,C_TRADE,From,NULL);
+               SendServerMessage(NULL,C_NONE,C_TRADE,From,NULL);*/
             }
          } else {
             sprintf(text,"Too late - %s has just left!",From->Attacked->Name);
@@ -1823,39 +1882,67 @@ int LoseBitch(struct PLAYER *Play,Inventory *Guns,Inventory *Drugs) {
 }
 
 void SetFightTimeout(struct PLAYER *Play) {
-   if (FightTimeout) Play->Timeout=time(NULL)+(time_t)FightTimeout;
-   else Play->Timeout=0;
+   if (FightTimeout) Play->FightTimeout=time(NULL)+(time_t)FightTimeout;
+   else Play->FightTimeout=0;
 }
 
 void ClearFightTimeout(struct PLAYER *Play) {
-   Play->Timeout=0;
+   Play->FightTimeout=0;
+}
+
+int AddTimeout(time_t timeout,time_t timenow,int *mintime) {
+   if (timeout==0) return 0;
+   else if (timeout<=timenow) return 1;
+   else {
+      *mintime=timeout-timenow;
+      return 0;
+   }
 }
 
 int GetMinimumTimeout(struct PLAYER *First) {
    struct PLAYER *tmp;
-   time_t timenow;
    int mintime=-1;
+   time_t timenow;
+
    timenow=time(NULL);
+   if (AddTimeout(MetaTimeout,timenow,&mintime)) return 0;
    for (tmp=First;tmp;tmp=tmp->Next) {
-      if (tmp->Timeout!=0) {
-/*         printf("Timeout for player %s is %ld, time is %ld\n",tmp->Name,
-                tmp->Timeout,timenow); */
-         if (tmp->Timeout<=timenow) return 0;
-         if (mintime==-1 || tmp->Timeout-timenow < mintime) {
-            mintime=tmp->Timeout-timenow;
-         }
-      }
+      if (AddTimeout(tmp->FightTimeout,timenow,&mintime) ||
+          AddTimeout(tmp->IdleTimeout,timenow,&mintime) ||
+          AddTimeout(tmp->ConnectTimeout,timenow,&mintime)) return 0;
    }
    return mintime;
 }
 
-void HandleTimeouts(struct PLAYER *First) {
-   struct PLAYER *tmp;
-   time_t timenow;
+void HandleTimeouts(struct PLAYER **FirstPt) {
+   struct PLAYER *tmp,*tmp2,*First;
    char text[BUFLEN];
+   time_t timenow;
+
+   First=*FirstPt;
    timenow=time(NULL);
-   for (tmp=First;tmp;tmp=tmp->Next) {
-      if (tmp->Timeout!=0 && tmp->Timeout<=timenow) {
+   if (MetaTimeout!=0 && MetaTimeout<=timenow) {
+      MetaTimeout=0;
+      RegisterWithMetaServer(TRUE,FALSE);
+   }
+   tmp=First;
+   while (tmp) {
+      tmp2=tmp->Next;
+      if (tmp->IdleTimeout!=0 && tmp->IdleTimeout<=timenow) {
+         tmp->IdleTimeout=0;
+         LOGF("Player removed due to idle timeout\n");
+         SendPrintMessage(NULL,C_NONE,tmp,"Disconnected due to idle timeout");
+         ClientLeftServer(tmp);
+         shutdown(tmp->fd,SD_SEND);
+/* Make sure they do actually disconnect, eventually! */
+         if (ConnectTimeout) {
+            tmp->ConnectTimeout=time(NULL)+(time_t)ConnectTimeout;
+         }
+      } else if (tmp->ConnectTimeout!=0 && tmp->ConnectTimeout<=timenow) {
+         tmp->ConnectTimeout=0;
+         LOGF("Player removed due to connect timeout\n");
+         RemovePlayer(tmp,FirstPt);
+      } else if (tmp->FightTimeout!=0 && tmp->FightTimeout<=timenow) {
          ClearFightTimeout(tmp);
          if (tmp->EventNum==E_WAITATTACK) {
             tmp->EventNum=E_FREEFORALL;
@@ -1867,5 +1954,6 @@ void HandleTimeouts(struct PLAYER *First) {
             SendServerMessage(tmp,C_NONE,C_FIGHTPRINT,tmp->Attacked,text);
          }
       }
+      tmp=tmp2;
    }
 }
